/**************************************************************************
   zlib.h      include file for zlib functions
   Free software licensed under GNU General Public License v.2
***************************************************************************/

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>
#include <setjmp.h>
#include <pthread.h>
#include <errno.h>
#include <signal.h>

#define  int8   char                                                       //  integer types
#define  int16  short
#define  int32  long
#define  int64  long long

#define  mutex          pthread_mutex_t                                    //  abbreviations
#define  mutex_init     pthread_mutex_init
#define  mutex_lock     pthread_mutex_lock
#define  mutex_unlock   pthread_mutex_unlock
#define  mutex_destroy  pthread_mutex_destroy

#define  IDCC  50                                                          //  max. symbol/value cc
#define  MAXFCC 1000                                                       //  max. pathname/filename cc
#define  null   NULL
#define  true   1
#define  false  0

//  macro, call a function within caller's context ------------------------

#define GOSUB_DATA(maxlevels)                                     \
int         gosub_stat;                                           \
int         gosub_level = 0;                                      \
int         gosub_maxlevel = (maxlevels);                         \
jmp_buf     gosub_jb[(maxlevels)];                                \

#define GOSUB(label)                                              \
{                                                                 \
   gosub_level++;                                                 \
   if (gosub_level >= gosub_maxlevel)                             \
            appcrash("exceeded GOSUB levels");                    \
   gosub_stat = setjmp(gosub_jb[gosub_level]);                    \
   if (! gosub_stat) goto label;                                  \
   gosub_level--;                                                 \
}                                                                 \

#define RETURN  longjmp(gosub_jb[gosub_level],1);

//  system functions ------------------------------------------------------

char * syserrText(int err = 0);                                            //  get sys error text (err > 0, else errno)
void appcrash(char * pMess, ... );                                         //  fatal exit - works like printf
void apppause();                                                           //  output message and wait for user
void apppause(char * pMess, ... );                                         //    - works like printf
void start_timer(timeval *time0);                                          //  start a timer
double get_timer(timeval *time0);                                          //  get elapsed time in seconds
double CPUtime();                                                          //  get process CPU time used
void start_CPUtimer(double *time0);                                        //  start a process CPU timer
double get_CPUtimer(double *time0);                                        //  get elapsed CPU time, seconds
void synch_threads(int NT = 0);                                            //  synchronize NT threads
void zsleep(double dsecs);                                                 //  sleep specified seconds
void * zmalloc(size_t bytes);                                              //  malloc() with auto crash on failure
int createProc(char * command);                                            //  create subprocess
int getProcOutput(char * buff, int bcc);                                   //  get output record from subprocess
int signalProc(char * pname, char * signal);                               //  pause/resume/kill subprocess
char * fgets_trim(char * buff, int maxcc, FILE *, int bf = 0);             //  fgets + trim trailing \n \r (opt. blanks)
int lrandz(int64 * seed);                                                  //  returns 0 to 0x7fffffff
double drandz(int64 * seed);                                               //  returns 0.0 to 0.99999...

//  string macros and functions -------------------------------------------

#define strEqu(str1,str2) (! strcmp((str1),(str2)))                        //  TRUE if strings compare equal
#define strNeq(str1,str2) (strcmp((str1),(str2)))                          //  TRUE if strings compare not-equal
#define strnEqu(str1,str2,cc) (! strncmp((str1),(str2),(cc)))              //  TRUE if strings compare equal
#define strnNeq(str1,str2,cc) (strncmp((str1),(str2),(cc)))                //  TRUE if strings compare not-equal

char * strParse(const char * string, char delim, int & ftf);               //  get next delimited field from string
int strHash(const char *, const int max);                                  //  string --> random 0 to max-1
void strncpy0(char * dest, const char * src, int cc);                      //  strncpy + dest[cc-1] = 0
void strnPad(char * dest, const char * src, int cc);                       //  strncpy with blank padding to cc
int strTrim(char * dest, const char * src);                                //  remove trailing blanks
int strTrim(char * string);                                                //  remove trailing blanks
int strCompress(char * dest, const char * sorc);                           //  copy & remove imbedded blanks
int strCompress(char * string);                                            //  remove imbedded blanks
int strncatv(char * dest, int maxcc, const char * source, ...);            //  catenate strings (last = 0L) 
int strcmpv(const char * string, ...);                                     //  compare string to N other strings
void strToUpper(char * dest, const char * source);                         //  move and conv. string to upper case
void strToUpper(char * string);                                            //  conv. string to upper case
void strToLower(char * dest, const char * source);                         //  move and conv. string to lower case
void strToLower(char * string);                                            //  conv. string to lower case
int repl_1str(char *strin, char *strout, char *ssin, char *ssout);         //  copy string and replace 1 substring
int repl_Nstrs(char *strin, char* strout, ...);                            //  copy string and replace N substrings
void strncpyx(char * dest, const char * source, int scc);                  //  conv. string to hex format
void StripZeros(char * pNum);                                              //  1.230000E+12 --> 1.23E+12

//  number conversion -----------------------------------------------------

int convSI (const char * string, int & inum);                              //  string to int
int convSI (const char *, int &, int low, int hi);                         //  (with low/high limit checking)
int convSD (const char * string, double & dnum);                           //  string to double
int convSD (const char *, double &, double low, double hi);                //  (with low/high limit checking)
int convIS (int in, char * out, int & cc);                                 //  int to string, returned length
int convDS (double in, int prec, char * out, int & cc);                    //  double to string, digits of precision

//  wildcard routines -----------------------------------------------------

int MatchWild(const char * wildstr, const char * str);                     //  wildcard string match (0)
const char * SearchWild(const char *wpath, int &flag);                     //  wildcard file search

//  search and sort functions ---------------------------------------------

int bsearch(int element, int nn, int list[]);                              //  binary search sorted list[nn]
void HeapSort(int vv[], int nn);                                           //  Heap Sort - integers
void HeapSort(float vv[], int nn);                                         //  Heap Sort - floats
void HeapSort(double vv[], int nn);                                        //  Heap Sort - doubles
void HeapSort(char * vv[], int nn);                                        //  Heap Sort - pointers to strings
typedef int HeapSortUcomp(char * rec1, char * rec2);                       //  compare function: return -1/0/+1 if rec1 </=/> rec2
void HeapSort(char * recs, int RL, int NR, HeapSortUcomp);                 //  Heap Sort - sort records by compare function
int MemSort(char * RECS, int RL, int NR, int KEYS[][3], int NK);           //  memory sort, records with multiple keys

//  variable length string class ------------------------------------------

class xstring
{
      static int  tcount;                                                  //  total xstring count
      static int  tmem;                                                    //  total memory used
      int         wmi;                                                     //  internal ID
      int         xcc;                                                     //  actal cc (excl. NULL)
      int         xmem;                                                    //  memory allocated cc
      char      * xpp;                                                     //  memory pointer
   public:
      xstring(int cc = 0);                                                 //  default constructor
      xstring(const char * );                                              //  string constructor
      xstring(const xstring &);                                            //  copy constructor
      ~xstring();                                                          //  destructor
      operator const char * () const { return xpp; }                       //  conversion operator (const char *)
      xstring operator= (const xstring &);                                 //  operator =
      xstring operator= (const char *);                                    //  operator =
      friend xstring operator+ (const xstring &, const xstring &);         //  operator +
      friend xstring operator+ (const xstring &, const char *);            //  operator +
      friend xstring operator+ (const char *, const xstring &);            //  operator +
      void insert(int pos, const char * string, int cc = 0);               //  insert substring at position (expand)
      void overlay(int pos, const char * string, int cc = 0);              //  overlay substring (possibly expand)
      static void getStats(int & tcount2, int & tmem2);                    //  get statistics
      void validate() const;                                               //  verify integrity
      int getcc() const { return xcc; }                                    //  return string length
};

//  vector (array) of xstring ---------------------------------------------

class Vxstring
{
   int        nd;                                                          //  count
   xstring  * pdata;                                                       //  xstring[nd]
public:
   Vxstring(int = 0);                                                      //  constructor
   ~Vxstring();                                                            //  destructor
   Vxstring(const Vxstring &);                                             //  copy constructor
   Vxstring operator= (const Vxstring &);                                  //  operator =
   xstring & operator[] (int);                                             //  operator []
   const xstring & operator[] (int) const;                                 //  operator [] (const)
   int search(const char * string);                                        //  find element in unsorted Vxstring
   int bsearch(const char * string);                                       //  find element in sorted Vxstring
   int sort(int nkeys, int keys[][3]);                                     //  sort elements by designated subfields
   int sort(int pos = 0, int cc = 0);                                      //  sort elements by 1 subfield (cc 0 = all)
   int getCount() const { return nd; }                                     //  get current count
};

//  hash table class ------------------------------------------------------

class HashTab
{
   static int trys1;                                                       //  insert trys
   static int trys2;                                                       //  find/delete trys
   int     cap;                                                            //  table capacity
   int     count;                                                          //  strings contained
   int     cc;                                                             //  string length
   char  * table;                                                          //  table[cc][cap]
public:
   HashTab(int cc, int cap);                                               //  constructor
   ~HashTab();                                                             //  destructor
   int Add(const char * string);                                           //  add a new string
   int Del(const char * string);                                           //  delete a string
   int Find(const char * string);                                          //  find a string
   int GetCount() { return count; };                                       //  get string count
   int GetNext(int & first, char * string);                                //  get first/next string
   int Dump();                                                             //  dump hash table
};

//  Queue class, FIFO, LIFO or mixed  -------------------------------------

class  Queue
{
   char        wmi[8];
   Vxstring  * vd;                                                         //  vector of xstrings
   mutex       qmutex;                                                     //  for multi-thread access
   int         qcap;                                                       //  queue capacity
   int         qcount;                                                     //  curr. queue count
   int         ent1;                                                       //  first entry pointer
   int         entN;                                                       //  last entry pointer
   char      * lastent;                                                    //  last entry retrieved
   int         lcc;                                                        //  last entry cc
private:
   void lock();                                                            //  auto locking and unlocking
   void unlock();                                                          //  (for multi-thread access)
public:
   Queue(int cap);                                                         //  create queue with capacity
   ~Queue();                                                               //  destroy queue
   int getCount();                                                         //  get current entry count
   int push(const xstring * entry, double secs);                           //  add new entry with max. wait time
   xstring * pop1();                                                       //  get 1st entry (oldest)
   xstring * popN();                                                       //  get Nth entry (newest)
};

//  Tree class  -----------------------------------------------------------
//       put(): dd is data length to store
//       get(): dd is max. data length to retrieve
//              actual length is returned, = 0 if not found

class  Tree
{
   int         wmi;                                                        //  for ID checking
   char       *tname;                                                      //  tree name
   int         tmem;                                                       //  tree data memory
   void       *tdata;                                                      //  tree data[tmem]
   int         nsub;                                                       //  no. sub-nodes (Trees)
   Tree      **psub;                                                       //  pointer to sub-nodes
public:
   Tree(char * name);                                                      //  create Tree
   ~Tree();                                                                //  destroy Tree
   int put(void * data, int dd, char * nodes[], int nn);                   //  put data by node names[]
   int put(void * data, int dd, int nodes[], int nn);                      //  put data by node numbers[]
   int get(void * data, int dd, char * nodes[], int nn);                   //  get data by node names[]
   int get(void * data, int dd, int nodes[], int nn);                      //  get data by node numbers[]
   void stats(int nnodes[], int ndata[]);                                  //  get nodes and data per level
   void dump(int level = 0);                                               //  diagnostic
private:
   Tree * find(char * nodes[], int nn);                                    //  find a sub-node by names[]
   Tree * find(int nodes[], int nn);                                       //  find a sub-node by numbers[]
   Tree * make(char * nodes[], int nn);                                    //  find/create a sub-node by names[]
   Tree * make(int nodes[], int nn);                                       //  find/create a sub-node by numbers[]
};


/**************************************************************************
   GTK utility functions
***************************************************************************/

#include <gtk/gtk.h>

//   macros to build menus and toolbars with less code

#define G_SIGNAL(window,event,func,arg)                                                \
   g_signal_connect(G_OBJECT(window),event,G_CALLBACK(func),(void *) arg);

#define create_menubar(mbname)                                                         \
   GtkWidget * mbname = gtk_menu_bar_new();                                            \

#define add_menubar_item(mbname,itname,label,func)                                     \
   GtkWidget * itname =  gtk_menu_item_new_with_label(label);                          \
   gtk_menu_shell_append(GTK_MENU_SHELL(mbname), itname);                              \
   G_SIGNAL(itname,"activate",func,label)

#define connect_submenu(itname,submenu)                                                \
   GtkWidget * submenu = gtk_menu_new();                                               \
   gtk_menu_item_set_submenu(GTK_MENU_ITEM(itname),submenu);

#define add_submenu_item(submenu,itname,label,func);                                   \
   GtkWidget * itname = gtk_menu_item_new_with_label(label);                           \
   gtk_menu_shell_append(GTK_MENU_SHELL(submenu),itname);                              \
   G_SIGNAL(itname,"activate",func,label)

#define create_toolbar(tbname)                                                         \
   GtkWidget * tbname = gtk_toolbar_new();                                             \
   GtkTooltips * tbname##tips = gtk_tooltips_new();

#define add_toolbar_button(tbname,itname,label,tiptext,func)                           \
   GtkToolItem * itname = gtk_tool_button_new(NULL,label);                             \
   gtk_toolbar_insert(GTK_TOOLBAR(tbname),GTK_TOOL_ITEM(itname),-1);                   \
   gtk_tool_item_set_tooltip(itname,tbname##tips,tiptext,"");                          \
   G_SIGNAL(itname,"clicked",func,label)

#define add_toolbar_stock_item(tbname,itname,label,tiptext,stockname,func)             \
   GtkToolItem * itname = gtk_tool_button_new_from_stock(GTK_STOCK_##stockname);       \
   gtk_tool_button_set_label(GTK_TOOL_BUTTON(itname),label);                           \
   gtk_toolbar_insert(GTK_TOOLBAR(tbname),GTK_TOOL_ITEM(itname),-1);                   \
   gtk_tool_item_set_tooltip(itname,tbname##tips,tiptext,"");                          \
   G_SIGNAL(itname,"clicked",func,label)

/**************************************************************************/

#define  PDF_VIEWERS "acroread -geometry 400x500","gpdf","evince"

void zappcrash(char * pMessage, ...);                                      //  crash with popup message

void zlockInit();                                                          //  for threaded applications
void zlock();  
void zunlock();

void zmainloop();                                                          //  do main loop, process menu events

void wpauseOK(char * pMess, ... );                                         //  display message, wait for OK
int wpauseYN(char * pMess, ... );                                          //  display message, wait for YES or NO

const char * dialogText(const char * title, const char * inittext);        //  get text (1-2 lines) from user

void wprintx(GtkWidget *textWin, int line, char *mess);                    //  write text to line +-N
void wprintf(GtkWidget *textWin, int line, char *mess, ... );              //  "printf" version
void wprintf(GtkWidget *textWin, char *mess, ... );                        //  "printf" with scrolling output
void wclear(GtkWidget * textWin);                                          //  clear window

char * wscanf(GtkWidget * textWin, int & ftf);                             //  get text lines from edit widget

int wfiledump(GtkWidget * textWin, char * filespec);                       //  dump text window into file
void wfilesave(GtkWidget * textWin);                                       //  wfiledump() via file-chooser dialog
void wprintp(GtkWidget * textWin);                                         //  print text window to default printer

char * zfile_open(char * title, char * initfile);                          //  file open dialog
char * zfile_save(char * title, char * initfile);                          //  file save dialog

/**************************************************************************/

//  parameter management functions

int initParmlist(int max);                                                 //  initz. parameter list
int loadParms();                                                           //  load parameters from a file
int editParms(int addp = 0);                                               //  parameter editor
int saveParms();                                                           //  save parameters to a file
int setParm(const char * parmname, double parmval);                        //  set parameter value
double getParm(const char * parmname);                                     //  get parameter value
char * getParm(int Nth);                                                   //  get Nth parameter name (0-based)

/**************************************************************************/

//   functions to implement dialogs with less complexity
//   widgets supported: dialog, hbox, vbox, hsep, vsep, frame, scrwin, label, 
//                      entry, edit, radio, check, button, togbutt, spin, 
//                      combo, hscale, vscale
   
typedef struct zdialog  {
         void        *eventCB;            //  widget event user callback function
         void        *complCB;            //  dialog completion user callback function
         int         zstat;               //  dialog status (from completion button)
      struct  {                           //  widget data structure
         const char  *type;               //  dialog, hbox, vbox, label, entry ...
         const char  *name;               //  widget name
         const char  *pname;              //  parent (container) name
         const char  *data;               //  widget data, initial / returned
         int         scc;                 //  entry widget: text cc (space)
         int         homog;               //  hbox/vbox: equal spacing
         int         expand;              //  expandable
         int         space;               //  extra space
         GtkWidget   *widget;             //  GTK widget pointer
      }  
      widget[100];                        //  zdialog widgets (eof = type = 0)
};

zdialog *zdialog_new(const char *title, ...);                              //  create a zdialog with opt. buttons

int zdialog_add_widget(zdialog *zd,                                        //  add widget to zdialog
      const char *type, const char *name, const char *pname,               //  3 required args
      const char *data = 0, int scc = 0, int homog = 0,                    //  5 optional args
      int expand = 0, int space = 0);

int zdialog_resize(zdialog *zd, int width, int height);                    //  set size > widget sizes
int zdialog_set_data(zdialog *zd, const char *name, const char *data);     //  set data (entry, edit, spin ...)
int zdialog_add_data(zdialog *zd, const char *name, const char *sdata);    //  add item to combo list

typedef int zdialog_event(zdialog *zd, int widget);                        //  widget event callback function
typedef int zdialog_compl(zdialog *zd, int stat);                          //  dialog completion callback function
int zdialog_run(zdialog *zd, zdialog_event func = null,                    //  run dialog with optional callbacks
                             zdialog_compl func = null);                   //  (completion func >> non-modal dialog)

int zdialog_destroy(zdialog *zd);                                          //  destroy dialog (zdialog caller resp.)

GtkWidget * zdialog_widget(zdialog *zd, const char *name);                 //  GTK widget from zdialog widget name

int zdialog_stuff(zdialog *zd, const char *name, const char *sdata);       //  stuff string data into widget
int zdialog_stuff(zdialog *zd, const char *name, int  idata);              //  stuff int data
int zdialog_stuff(zdialog *zd, const char *name, double ddata);            //  stuff double data

const char * zdialog_get_data(zdialog *zd, const char *name);              //  retrieve zdialog widget data
int zdialog_fetch(zdialog *zd, const char *name, char *sdata, int maxcc);  //  get string data from widget
int zdialog_fetch(zdialog *zd, const char *name, int  &idata);             //  get int data
int zdialog_fetch(zdialog *zd, const char *name, double &ddata);           //  get double data

